/**
 * \file
 * Class XALM.
 */

#ifndef XALM_H
#define XALM_H

#include "arrays.h"

using namespace gelibspace;

namespace xalm {

/// XALM mody; urcuji typ kontroly delky kroku.
enum xalm_ControlType {
  xalm_hpc_off = 0,  ///<  Full ALM with quadratic constrain and all dofs.
  xalm_hpc_on,       ///<  Full ALM with quadratic constrain, taking into account only selected dofs.
  xalml_hpc,         ///<  Linearized ALM (only displacements), taking into account only selected dofs with given weight.
};

/// Strategie výpočtu matice tuhosti, převzato z Newton-Raphson metody iteračního výpočtu.
enum xalm_NR_ModeType {
  xalm_modifiedNRM,  ///< Modifikovaná NR metoda (defaultní hodnota) - matice se počítá jen na začátku každého zatěžovacího kroku.
  xalm_fullNRM,      ///< Plná NR metoda - matice se během zatěžovacího kroku počítá při každé iteraci.
  xalm_accelNRM,     ///< Akcelerovaná NR metoda - matice se počítá jen na začátku každého n-tého zatěžovacího kroku, n si určí uživatel.
};

#define XALM_RESET_STEP_REDUCE 0.25
#define XALM_MAX_RESTARTS 4
#define XALM_DEFAULT_NRM_TICKS 2
#define XALM_MAX_REL_ERROR_BOUND 1.e10

#define XALM_SMALL_NUM 1.e-20
#define XALM_SMALL_ERROR_NUM 1.e-6

/**
 *  Hlavní, a jedíná, třída knihovny XALM.
 *  \nosubgrouping
 *
 *  EXternal cylindrical Arc-Length Method algorithm.
 *
 * Perform solution of non-linear problem in the form
 * Kt * deltaR = g
 * where g is defined as g = g(Lambda,X) = Lambda* R - F(X).
 * During iteration process, we want g became zero vector.
 *
 *
 */
class XALM
{
  // *****************************************************************************************
  // *************************             VEŘEJNÉ METODY             ************************
  // *****************************************************************************************
  /**
     @name Veřejné metody
     
     Veřejné metody slouží pro spuštění výpočtu a pro předávání výsledků uživatelovi.
  */
  // @{
 public:
  /// Constructor.
  XALM (void);
  /// Destructor.
  virtual ~XALM (void);
  
  /**
   * Funkce solve je hlavní, a jediná, výkonná funkce knihovny XALM.
   */
  int solve (void);
  
  /// Funkce vrátí konstantní ukazatel na vektor celkových posunů.
  const double* give_totalDisplacement (void) const { return totalDisplacement.a; }
  /// Funkce vrátí loadLevel - dosažený stupeň přírůstkového zatížení.
  double give_loadLevel (void) const { return loadLevel; }
  
  // @}
  
  
  
  // *****************************************************************************************
  // *************************       SOUKROMÉ METODY A ATRIBUTY       ************************
  // *****************************************************************************************
  /**
     @name Soukromé metody a atributy
     
     Soukromé metody a atributy nejsou přístupné pro používání z vnějšku třídy ani z třídy zděděné a slouží pro vlastní výpočet.
     Pokud jsou soukromé atributy ve zděděné třídě potřebné, přistupuje se k nim pomocí chráněných metod.
  */
  // @{
 private:
  int step;                          ///<  Aktuální krok.
  double loadLevel;                  ///<  Dosažený stupeň přírůstkového zatížení.
  Dvctr totalDisplacement;           ///<  Vektor celkových posunutí.
  Dvctr incrementOfDisplacement;     ///<  Vektor přírůstku posunutí.
  Dvctr internalForces;              ///<  Vektor vnitřních sil v uzlech.
  int status;                        ///<  0 - nedopočítáno, 1- spočítáno, 2 - počítáno s problémy, změněn krok
  xalm_NR_ModeType xalm_NR_OldMode;  ///<  Záloha proměnné xalm_NR_Mode.
  int xalm_NR_ModeTick;              ///<  Počet kroků zbývajících do konce dočasné změny iteračního schématu.
  
  ///
  int computeDeltaLambda (double &deltaLambda, const Dvctr &dX, const Dvctr &deltaXt,
			  const Dvctr &deltaX_, const Dvctr &R, double RR, double eta,
			  double deltaL, double DeltaLambda0, int neq);
  
  void solve_step(Dvctr *R, Dvctr *R0, Dvctr *X, Dvctr *dX, Dvctr *F,  long step);
  //double giveCurrentStepLength() { return deltaL; }
  //void setStepLength(double s) { deltaL = s; }
  
  /// Funkce pro nastavení soukromých atributů a kontrolu konzistence zadání.
  void initialize_and_check_consistency (void);
  /// Kontrola konvergence.
  bool checkConvergence (const Dvctr &R, const Dvctr *R0, const Dvctr &F, const Dvctr &X, const Dvctr &ddX,
			 double Lambda, double RR0, double RR, double drProduct, int nite, bool &errorOutOfRange);
  
  
  // Docasne nepouzivane. -  experimentalni podpora
  // Dvctr internalForcesEBENorm; pocita se pri vypoctu vektoru vnitrnich sil
  //int xalm_hpc_init;                /// HyperPlaneControl.
  xalm_ControlType xalm_Control;
  Dvctr xalm_HPCWeights;
  Lvctr xalm_HPCIndirectDofMask;    /// Array containing equation numbers of dofs under indirect control.
  Lvctr xalm_HPCDmanDofSrcArray;    /// Input array containing dofmanagers and corresponding dof numbers under indirect control.
  Dvctr xalm_HPCDmanWeightSrcArray; /// Input array of dofman weights (for hpcmode 2).
  
  // @}
  
  
  
  // *****************************************************************************************
  // *************************       CHRÁNĚNÉ METODY A ATRIBUTY       ************************
  // *****************************************************************************************
  /**
     @name Chráněné metody a atributy
     
     Chráněné metody a atributy slouží pro zadávání počátečních požadavků na
     výpočet a pro předávání potřebných dat mezi MKP programem a knihovnou XALM
     během vlastního iteračního výpočtu.
     
     V třídě XALM jsou proto metody deklarovány jako čistě virtuální (pure virtual) a definovány jsou až v třídě zděděné.
     Je zřejmé, že metody jsou implementačně závislé.
     Proto zděděná třída nemůže být součástí knihovny XALM, ale musí být vytvořena pro každou
     implementaci zvlášť a je její součástí.
     Tvoří tak fakticky "rozhraní" popisované v druhé kapitole.
  */
  // @{
 protected:
  
  //* *** Chráněné atributy ***
  
  // Základní proměnné, které řídí běh ALM metody. Jsou volené "uživatelem".
  // Tyto atributy se musí nastavit na začátku výpočtu ve funkci "initialize" ve zděděné třídě rozhraní.
  xalm_NR_ModeType xalm_NR_Mode;  ///<  Proměnná určující strategii, kdy se bude počítat tečná matice tuhosti soustavy.
  int xalm_MANRMSteps;            ///<  Počet kroků, po kterých se má znovu počítat matice, pouze pokud xalm_NR_Mode == xalm_accelNRM.
  long   nsteps;                  ///<  Počet zatěžovacích kroků.
  double minStepLength;           ///<  Minimalní délka kroku.
  double maxStepLength;           ///<  Maximalní délka kroku.
  double initialStepLength;       ///<  Počáteční délka kroku.
  int nsMin;                      ///<  Minimální  počet kroků dorovnání nerovnováhy v jednom iteračním přitěžovacím kroku.
  int nsReq;                      ///<  Požadovaný počet kroků dorovnání nerovnováhy v jednom iteračním přitěžovacím kroku.
  int nsMax;                      ///<  Maximální  počet kroků dorovnání nerovnováhy v jednom iteračním přitěžovacím kroku.
  double Psi;                     ///<  Parametr kontroly kroku. Pokud je rovno 0, tak se jedná o kontrolu přírůstkem posunutí, pokud je rovno nekonečnu, tak se jedná o kontrolu přírůstkem zatížení.
  int verbose;                    ///<  Pokud je větší než nule - budou vypisovány informace.
  double rtolf;                   ///<  Tolerance relativní chyby nevyrovnaných sil.
  double rtold;                   ///<  Tolerance relativní chyby nevyrovnaných posunů.
  
  // Tyto argumenty se nevolí, ale závisí na zadání právě počítané úlohy.
  // Tyto atributy se musí nastavit na začátku výpočtu ve funkci "initialize" ve zděděné třídě rozhraní.
  long neq;                       ///<  Pocet rovnic v matici soustavy = pocet neznamych.
  Dvctr incrementalLoadVector;    ///<  Vektor prirustkoveho zatizeni, meni se se stupnem lambda.
  Dvctr initialLoadVector;        ///<  Vektor pocatecniho zatizeni, nemeni se behem vypoctu, je aplikovan cely.
  
  
  //* *** Chráněné metody ***
  
  // Metody pro přístup k aktuálním stavovým hodnotám.
  /// Funkce vrací číslo aktuálního zatěžovacího kroku.
  int give_step (void) { return step; }
  
  
  //* *** Chráněné metody s návodem na implementaci v zděděné třídě: */
  
  /// Funkce nemá parametry a nic nevrací. Má za úkol inicializovat chráněné atributy.
  /// Atributy typu ukazatel na pole musí být alokovány uvnitř této funkce. Dealokovány jsou v destruktoru třídy XALM.
  virtual void initialize (void) = 0;
  
  /// Ve funkci je možné po každém provedeném iteračním kroku vykonat požadované úkony.
  /// Např. vypsat aktuání hodnoty, zastavit iteraci, pokud je dosaženo požadovaných výsledků, atd.
  virtual void update_step (void) = 0;
  
  ///
  virtual void update_stiffness_matrix (const Dvctr *X) = 0;
  /// 
  virtual void update_internal_forces (Dvctr *internalForces, const Dvctr *X) = 0;
  ///
  virtual void lineq_solve (Dvctr *X, const Dvctr *R) = 0;
  
  // @}
};

} // end namespace xalm

#endif // XALM_H
